/*
 * report.c - calculation and command line reporting
 *
 * $Id$
 *
 * Copyright (c) 2001, The Colocation Corporation. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.  Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.  Neither the name of the The
 * Colocation Corporation nor the names of its contributors may be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * This software was originally written under grants from the Colocation
 * Corporation. To find the most recent versions of the software, please visit
 * http://www.loligo.com/ocoloc  To contribute patches or for other questions
 * please contact jtodd@loligo.com
 *
 * To find out more about the Colocation Corporation, see http://www.coloco.com/
 *
 */

#include <limits.h>
#include "ocoloc.h"

unsigned long
delta(unsigned long from, unsigned long amt)
{
   if (from < amt) {
      unsigned long long ull = ULONG_MAX + 1 + from - amt;
      return ((unsigned long) ull & ULONG_MAX);
   } else
      return (from - amt);
}

void
apply_units(Options * opt, Stats * s)
{
    s->units = opt->unit_name;

    if (opt->units == bytes)
	return;

    if (opt->units == bits) {

	unsigned long multiple = 8;

	s->min_thru *= 8;
	s->max_thru *= 8;
	s->wgt_in *= 8;
	s->wgt_out *= 8;
	s->wgt_agg *= 8;
	mpz_mul_ui(s->avg_in, s->avg_in, multiple);
	mpz_mul_ui(s->avg_out, s->avg_out, multiple);
	mpz_mul_ui(s->ttl_in, s->ttl_in, multiple);
	mpz_mul_ui(s->ttl_out, s->ttl_out, multiple);
	return;
    }
    else if (opt->units == kilobytes) {

	unsigned long divisor = 1024;

	s->min_thru /= 1024;
	s->max_thru /= 1024;
	s->wgt_in /= 1024;
	s->wgt_out /= 1024;
	s->wgt_agg /= 1024;
    mpz_cdiv_q_ui(s->avg_in, s->avg_in, divisor);
    mpz_cdiv_q_ui(s->avg_out, s->avg_out, divisor);
    mpz_cdiv_q_ui(s->ttl_in, s->ttl_in, divisor);
    mpz_cdiv_q_ui(s->ttl_out, s->ttl_out, divisor);
	return;
    }
    printf("unknown units\n");
}

void
write_stats(Stats * s)
{

    if (!s)
	BAIL("stats is null\n");

    printf("\n%16s: %s", "Host", s->hostname);
    if (s->hostcomment)
	printf(" (%s)", s->hostcomment);

    printf("\n%16s: %s\n", "Interface", s->ifname);
    if (s->ifcomment)
	printf("%16s: %s\n", "Comment", s->ifcomment);

    printf("%16s: %s", "Start Time", ctime(&s->start));
    printf("%16s: %s", "End Time", ctime(&s->end));

    printf("%16s: %ld\n", "Min Thruput", s->min_thru);
    printf("%16s: %ld\n", "Max Thruput", s->max_thru);

    printf("     %2.0f%% in (%2s): %lu\n", 
        (s->pct * 100), s->units, s->wgt_in);
    printf("    %2.0f%% out (%2s): %lu\n", 
        (s->pct * 100), s->units, s->wgt_out);
    printf("    %2.0f%% agg (%2s): %lu", 
        (s->pct * 100), s->units, s->wgt_agg);

    printf("\n     Avg in (%2s): ", s->units);
    mpz_out_str(stdout, 10, s->avg_in);
    printf("\n    Avg out (%2s): ", s->units);
    mpz_out_str(stdout, 10, s->avg_out);

    printf("\n   Total in (%2s): ", s->units);
    mpz_out_str(stdout, 10, s->ttl_in);
    printf("\n  Total out (%2s): ", s->units);
    mpz_out_str(stdout, 10, s->ttl_out);

    printf("\n%16s: %d\n", "Samples", s->samples);

}

int
ulcmp(const void *a, const void *b)
{
   unsigned long  *x = (unsigned long *) a;
   unsigned long  *y = (unsigned long *) b;

   if (*x == *y)
      return 0;
   else if (*x < *y)
      return -1;
   else
      return 1;
}

void
report_interface(Defaults * def, Options * opt, DevList * host, IfList * port)
{

    Entry          *entry, *entry_list = 0, *last;
    Entry           thisEntry, prevEntry;
    char           *fix;
    char            fn[80], line[80];
    unsigned int    entries = 0, pct_smp;
    unsigned long  *in, *out, *ttl;
    unsigned long duration;

    port->stats.hostname = host->name;
    port->stats.hostcomment = host->comment;
    port->stats.ifname = port->name;
    port->stats.ifcomment = port->comment;

    port->stats.pct = (port->pct ? port->pct : def->percentage);

    if (opt->end < 1)
	opt->end = NOW;

    fix = port->name;
    while ((fix = strchr(fix, '/')))
	*fix = '_';

    sprintf(fn, "%s/%s/%s", def->datapath, host->name, port->name);

    if ((port->fp = fopen(fn, "r")) == NULL) {
	fprintf(stderr, "Can't open file %s\n", fn);
	return;
    }

    entry_list = 0;
    entries = 0;
    last = 0;
    memset(&thisEntry, 0, sizeof(Entry));
    memset(&prevEntry, 0, sizeof(Entry));

    while (!feof(port->fp)) {

	if (!(fgets(line, 80, port->fp)))
	    continue;

	if (sscanf(line, "%lu %u %lu %u %lu\n", &thisEntry.uptime,
			&thisEntry.in_time, &thisEntry.in, 
			&thisEntry.out_time, &thisEntry.out)
	    != 5) {
	    printf("Bad Line of data in:\n\t%s\n", fn);
	    return;
	}
	/* only handle our time range */
	if (thisEntry.in_time < opt->start || thisEntry.out_time > opt->end)
	    continue;

	/* if less than a second has passed (it happens, I guess)
           just skip this sample. */
        if ((thisEntry.in_time - prevEntry.in_time == 0) ||
	    (thisEntry.out_time - prevEntry.out_time == 0))
	   continue;

	/* if we rebooted the switch or the tick timer looped,
	then don't count this new sample, as it will throw off
	the delta count. However, record it for the following
	sample */
	if (thisEntry.uptime < prevEntry.uptime)
           goto skip;

	if (prevEntry.in_time != 0) {
	    assert(entry = malloc(sizeof(Entry)));
	    entry->in = delta(thisEntry.in, prevEntry.in);
	    entry->in_time = thisEntry.in_time - prevEntry.in_time;
	    entry->out = delta(thisEntry.out, prevEntry.out);
	    entry->out_time = thisEntry.out_time - prevEntry.out_time;
	    entry->next = (entry_list ? entry_list : 0);
	    entry_list = entry;
	    entries += 1;
	} else {
	    port->stats.start = thisEntry.in_time;
	}

skip:
	memcpy(&prevEntry, &thisEntry, sizeof(Entry));
    }

    fclose(port->fp);

    if (!entry_list) {
	fprintf(stderr, "No samples available for %s/%s\n",
		host->name, port->name);
	return;
    }
    port->stats.end = thisEntry.out_time;
    assert(in = calloc(entries, sizeof(unsigned long)));
    assert(out = calloc(entries, sizeof(unsigned long)));
    assert(ttl = calloc(entries, sizeof(unsigned long)));
    last = 0;

    entry = entry_list;
    entries = 0;
    while (entry) {
	in[entries] = entry->in / entry->in_time;
	out[entries] = entry->out / entry->out_time;
	ttl[entries] = (entry->in + entry->out) / entry->out_time;

        mpz_add_ui(port->stats.ttl_in, port->stats.ttl_in, entry->in);
        mpz_add_ui(port->stats.ttl_out, port->stats.ttl_out, entry->out);

	entry = entry->next;
	entries += 1;
    }

    port->stats.samples = entries;

    qsort((void *)in, entries, sizeof(unsigned long), ulcmp);
    qsort((void *)out, entries, sizeof(unsigned long), ulcmp);
    qsort((void *)ttl, entries, sizeof(unsigned long), ulcmp);

    port->stats.min_thru = ttl[0];
    port->stats.max_thru = ttl[port->stats.samples - 1];

    pct_smp = port->stats.samples * port->stats.pct;

    port->stats.wgt_in = in[pct_smp];
    port->stats.wgt_out = out[pct_smp];
    port->stats.wgt_agg = ttl[pct_smp];

    duration = port->stats.end - port->stats.start;
    mpz_cdiv_q_ui(port->stats.avg_in, port->stats.ttl_in, duration);
    mpz_cdiv_q_ui(port->stats.avg_out, port->stats.ttl_out, duration);

    apply_units(opt, &port->stats);

    port->stats.start = opt->start;
    port->stats.end = opt->end;

    if (opt->use_cgi)
	write_html_stats(&port->stats);
    else
	write_stats(&port->stats);

    free(in);
    free(out);
    free(ttl);

}

void
report_specific_interface(Defaults * def, Options * opt, DevList * tree)
{
    DevList        *d = tree;
    IfList         *i;

    while (d) {
	if (!strcasecmp(opt->device, d->name))
	    break;
	d = d->next;
    }
    if (!d)
	BAIL("Unknown host %s", opt->device);

    i = d->interface;
    while (i) {
	if (!strcasecmp(opt->interface, i->name))
	    break;
	i = i->next;
    }
    if (!i)
	BAIL("Unknown interface %s", opt->interface);

    report_interface(def, opt, d, i);
}

void
report_specific_host(Defaults * def, Options * opt, DevList * tree)
{
    DevList        *d = tree;
    IfList         *i;

    while (d) {
	if (!strcasecmp(opt->device, d->name))
	    break;
	d = d->next;
    }
    if (!d)
	BAIL("Unknown host %s", opt->device);

    i = d->interface;
    while (i) {
	report_interface(def, opt, d, i);
	i = i->next;
    }
}

void
report_data(Defaults * def, Options * opt, DevList * tree)
{
    DevList        *host = tree;
    IfList         *port;

    if (opt->device) {
	if (opt->interface)
	    report_specific_interface(def, opt, tree);
	else
	    report_specific_host(def, opt, tree);
	return;
    }
    while (host) {
	port = host->interface;
	while (port) {
	    report_interface(def, opt, host, port);
	    port = port->next;
	}
	host = host->next;
    }
}
